/*	spi.c

	Like pi.c, this program uses C with AFAPI to compute the value
	of pi.  However, it uses the signal-based replicated shared
	memory (RSM) mechanisms.  It is designed to be as comparable
	as possible to the other language examples in M. J.  Quinn,
	"Parallel Computing Theory And Practice," second edition,
	McGraw Hill, New York, 1994.  This code is most similar to the
	Sequent C code example given on page 108, although the AFAPI
	version is significantly shorter and simpler.  The differences
	center on the fact that our shared memory environment begins
	with all PEs active, whereas the Sequent C environment spawns
	groups of PEs for functions executed between barrier
	synchronizations.

	1996 by Hank Dietz and Tim Mattox
*/


#include <stdlib.h>
#include <stdio.h>
#include "afapi.h"


p_64f pi = 0.0;			/* Approximation to pi (shared) */
s_lock pilock = S_INITLOCK;	/* Lock for updates on pi */


main(int argc,
char **argv)
{
	register double intervals, width, localsum;
	register int i;

	/* check-in with AFAPI and enable signals */
	if (p_init()) exit(1);
	if (s_init()) exit(1);

	/* get the number of intervals */
	intervals = atoi(argv[1]);
	width = 1.0 / intervals;

	/* do the local computations */
	localsum = 0;
	for (i=IPROC; i<intervals; i+=NPROC) {
		register double x = (i + 0.5) * width;
		localsum += 4.0 / (1.0 + x * x);
	}
	localsum *= width;

	/* get permission to update pi, update it, and unlock again */
	s_acquire(&pilock);
	pi += localsum;
	s_update(&pi, sizeof(pi));
	s_release(&pilock);

	/* need to do a barrier to make sure all done */
	s_wait();

	/* have only the console PE print the result */
	if (IPROC == CPROC) {
		printf("Estimation of pi is %14.12lf\n", pi);
	}

	/* check-out */
	s_exit();
	p_exit();
	exit(0);
}
