/*	vote.c

	AFAPI voting and scheduling function definitions for
	TTL_PAPERS.

	July 1996 by H. Dietz
*/


#include "afapi.h"
#include "inline.h"

/*	p_first()...

	Return the IPROC of the first lowest-numbered PE with a yes
	vote.  The obvious way to do this is to collect a mask and
	then scan it; instead, we'll keep checking each piece as we
	get it, and return immediately when we find the first yes.
*/

static int p_first_tab[16] = {
/*	 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 */
	 4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
};

int
p_first(register p_1u f)
{
	register int i = 0;

	do {
		register int t = p_first_tab[
			p__nand((f && (IPROC >= i) && (IPROC < i+3)) ?
				(0xf ^ (1 << (IPROC - i))) :
				0xf)];

		if (t < 4) return(i + t);
	} while ((i += 4) < NPROC);

	return(NPROC);
}

/*	p_count()...

	Count how many voted yes.  Like p_first(), this doesn't bother
	to construct a bit mask.
*/

int p_count_tab[16] = {
/*	 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 */
	 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
};

int
p_count(register p_1u f)
{
	register int t = 0;
	register int i = 0;

	do {
		t += p_count_tab[
			p__nand((f && (IPROC >= i) && (IPROC < i+3)) ?
				(0xf ^ (1 << (IPROC - i))) :
				0xf)];
	} while ((i += 4) < NPROC);

	return(t);
}

/*	p_quantify()...

	Quantify the number of PEs voting yes; the return is 0, 1, 2
	(really >1), or NPROC.  The value is allowed to distinguish
	more than these counts, and actually returns a precise count
	in some implementations (e.g., here, for fewer than 9 PEs).
	Quantify is a surprisingly efficient operation using the PAPERS
	hardware to implement a systolic algorithm....
*/

int
p_quantify(register p_1u f)
{
#if NPROC < 9
	return(p_count(f));
#else
	register int mask =
#if NPROC < 17
		p__nand(f ? (0xf ^ IPROC) : 0xf);
#else
		/* This really only works for up to 256 PEs  ;-) */
		p_reduceOr8u(f ? IPROC : 0);
#endif
	if (f != 0) {
		mask = p__nand((IPROC == mask) | 2);
	} else {
		mask = p__nand(5);
	}

	if ((mask & 4) == 0) return(0);
	if ((mask & 2) == 0) return(NPROC);
	return(((mask & 1) == 0) ? 1 : 2);
#endif
}

/*	p_vote()...

	Builds a vote barrier bit vector using one p__waitbar() to
	determine an "upper bound" mask and then a sequence of Log2
	NPROC p__waitbar(), each of which determines who matched 1 bit
	of everybody's IPROC.
*/

inline p_bar
p_vote(register int pe)
{
	register p_bar bit = 1;
	register p_bar votemask = p__waitbar((pe >= 0) && (pe < NPROC));

	do {
		if (IPROC & bit) {
			votemask &= p__waitbar(pe & bit);
		} else {
			votemask &= ~p__waitbar(pe & bit);
		}
	} while ((bit <<= 1) & (NPROC-1));

	return(votemask);
}

/*	p_voteCount()...

	Unlike p_match(), doesn't bother with the masking stuff.
	Instead, just counts matches of pe to IPROC.

	This approach tends to waste more potential bits of bandwidth
	(due to 4-bit data path) than the technique used in p_vote(),
	but constructing and counting population of the bit mask is
	probably enough overhead to make that a moot issue.
*/

int
p_voteCount(register int pe)
{
	register int who = 0;
	register int count = 0;

	do {
#if	NPROC < 15
		/* 0..NPROC fits in a nybble... */
		if (IPROC == who) {
			/* Our turn to send data */
			p__put(pe);
		} else {
			/* We just get and compare data... */
			count += (p__get() == pe);
		}
#else
		/* A p_8u is enough for 255 PEs... */
		if (IPROC == who) {
			/* Our turn to send data */
			p_bcastPut8u(pe);
		} else {
			/* We just get and compare data... */
			count += (p_bcastGet8u() == pe);
		}
#endif
	} while (++who < NPROC);

	return(count);
}

/*	p_match()...
*/

#define	P_DECLARE(suffix) \
inline p_bar \
p_match##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_bar mask = 1; \
	register p_bar matchmask = (1 << IPROC); \
 \
	do { \
		if (IPROC == who) { \
			/* Our turn to send and store data */ \
			p_bcastPut##suffix(datum); \
		} else { \
			/* We just get and compare data... */ \
			if (datum == p_bcastGet##suffix()) { \
				matchmask |= mask; \
			} \
		} \
		mask <<= 1; \
	} while (++who < NPROC); \
 \
	return(matchmask); \
}

P_BETA

#undef	P_DECLARE

/*	p_matchCount()...

	Very much like p_match(), but doesn't bother with the mask;
	just counting is quicker.
*/

#define	P_DECLARE(suffix) \
int \
p_matchCount##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register int count = 1; /* We match ourself */ \
 \
	do { \
		if (IPROC == who) { \
			/* Our turn to send data */ \
			p_bcastPut##suffix(datum); \
		} else { \
			/* We just get and compare data... */ \
			count += (datum == p_bcastGet##suffix()); \
		} \
	} while (++who < NPROC); \
 \
	return(count); \
}

P_BETA

#undef	P_DECLARE

/*	p_rank()...
*/

#define	P_DECLARE(suffix) \
int \
p_rank##suffix(register p_##suffix datum) \
{ \
	register p_##suffix d; \
	register int who = 0; \
	register int rank = 0; \
 \
	do { \
		if (IPROC == who) { \
			/* Output our value */ \
			p_bcastPut##suffix(d = datum); \
		} else { \
			/* Increment our rank iff his value is bigger \
			   or his value equals our value and he has a \
			   lower value of IPROC. \
			*/ \
			register p_##suffix t = p_bcastGet##suffix(); \
 \
			rank += ((t < datum) || \
				 ((t == datum) && (who < IPROC))); \
		} \
	} while (++who < NPROC); \
 \
	return(rank); \
}

P_ALPHA

#undef	P_DECLARE
